/*
 * The SmartWeb Framework
 * Copyright (C) 2004-2006
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * For further informations on the SmartWeb Framework please visit
 *
 *                        http://smartweb.sourceforge.net
 */
package net.smartlab.web;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import net.smartlab.web.DataAccessObject.SearchInfo.Filter;
import net.smartlab.web.config.ConfigurationStrategy;
import net.smartlab.web.config.JNDIConfigurationStrategy;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.hibernate.Criteria;
import org.hibernate.HibernateException;
import org.hibernate.Query;
import org.hibernate.Session;
import org.hibernate.SessionFactory;
import org.hibernate.Transaction;
import org.hibernate.criterion.Expression;
import org.hibernate.criterion.Order;
import org.hibernate.criterion.Projections;
import org.hibernate.impl.CriteriaImpl;

/**
 * This class provides basic template for an Hibernate based implementation of
 * the DAO pattern.
 * 
 * @author rlogiacco@users.sourceforge.net
 */
public abstract class BusinessObjectFactory implements DataAccessObject {

	/**
	 * Provides logging capabilities to the factory.
	 */
	protected final Log logger = LogFactory.getLog(this.getClass());

	/**
	 * Stores the Hibernate Session instance available for this request.
	 */
	private static final ThreadLocal session = new ThreadLocal();

	/**
	 * Stores the Hibernate Transaction instance available for this request.
	 */
	private static final ThreadLocal transaction = new ThreadLocal();

	/**
	 * Caches a reference to the Hibernate Session Factory to avoid further
	 * lookups.
	 */
	protected SessionFactory factory;

	/**
	 * Strategy to retrieve the configuration file.
	 */
	private static ConfigurationStrategy strategy = new JNDIConfigurationStrategy();
	static {
		try {
			String strategy = System.getProperty("smartweb.conf.strategy");
			if (strategy != null) {
				BusinessObjectFactory.strategy = (ConfigurationStrategy)Class.forName(strategy).newInstance();
			} else {
				LogFactory.getLog(BusinessObjectFactory.class).warn(
						"No configuration found: falling back to default configuration");
			}
		} catch (Exception e) {
			LogFactory.getLog(BusinessObjectFactory.class).fatal("Error configuring SmartWeb", e);
		}
	}


	/**
	 * Changes the strategy used to configure the framework.
	 * 
	 * @param strategy an implementation of the
	 *        <code>ConfigurationStrategy</code> interface.
	 */
	public static void setConfigurationStrategy(ConfigurationStrategy strategy) {
		BusinessObjectFactory.strategy = strategy;
	}

	/**
	 * Closes the connection to the persistance layer.
	 * 
	 * @throws DAOException if the persistance layer generates any error while
	 *         performing the resource release.
	 */
	public static void close() throws DAOException {
		Session session = (Session)BusinessObjectFactory.session.get();
		if (session != null) {
			try {
				session.flush();
				session.close();
			} catch (HibernateException he) {
				throw new DAOException("errors.session.close", he);
			} finally {
				BusinessObjectFactory.session.set(null);
			}
		}
	}

	/**
	 * Protected default constructor to allow subclassing. Provides the startup
	 * initialization and the Hibernate Session Factory lookup. The Hibernate
	 * configuration file used must reside into the <code>META-INF</code>
	 * directory of the topmost packaging archive and should be named
	 * <code>smartweb.jar.hcf</code> or have the name of the JAR file
	 * containing the subclass followed by the <code>.hcf</code> suffix.
	 * 
	 * TODO example
	 */
	protected BusinessObjectFactory() {
		this.factory = strategy.getSessionFactory(this);
	}

	/**
	 * Returns the current Hibernate Session. If a Session was previously
	 * established during the same request the already established instance is
	 * returned otherwise a new instance is retriven from the Session Factory.
	 * 
	 * @return the current connection to the persistance layer or a new one if
	 *         not previously established.
	 * @throws DAOException if an error occurs trying to establish the
	 *         connection.
	 */
	protected Session current() throws DAOException {
		if (logger.isDebugEnabled()) {
			logger.debug("current() - start");
		}
		Session session = (Session)BusinessObjectFactory.session.get();
		if (session == null) {
			try {
				session = factory.openSession();
				BusinessObjectFactory.session.set(session);
			} catch (Exception e) {
				throw new DAOException("errors.session.config", e);
			}
		}
		return session;
	}

	/**
	 * @see net.smartlab.web.DataAccessObject#findByKey(java.io.Serializable)
	 */
	public Object findByKey(Serializable key) throws DAOException {
		if (logger.isDebugEnabled()) {
			logger.debug("findByKey(key = " + key + ") - start");
		}
		Serializable convertedKey = this.convertKey(key);
		if (convertedKey == null) {
			try {
				return this.getMappedClass().newInstance();
			} catch (Exception e) {
				logger.error("findByKey(key = " + key + ") - instantiation failed", e);
				throw new DAOException("errors.session.select", e);
			}
		} else {
			Session session = this.current();
			try {
				return session.get(this.getMappedClass(), convertedKey);
			} catch (HibernateException he) {
				logger.warn("findByKey(key = " + key + ") - deserialization failed", he);
				throw new UndefinedKeyException(key, this.getMappedClass(), he);
			}
		}
	}

	/**
	 * @see net.smartlab.web.DataAccessObject#remove(java.lang.Object)
	 */
	public void remove(Object object) throws DAOException {
		if (logger.isDebugEnabled()) {
			logger.debug("remove(object = " + object + ") - start");
		}
		try {
			this.current().delete(object);
		} catch (HibernateException he) {
			logger.warn("remove(object = " + object + ") - failed", he);
			throw new DAOException("errors.session.remove", he);
		}
	}

	/**
	 * @see net.smartlab.web.DataAccessObject#update(java.lang.Object)
	 */
	public void update(Object object) throws DAOException {
		if (logger.isDebugEnabled()) {
			logger.debug("update(object = " + object + ") - start");
		}
		try {
			this.current().saveOrUpdate(object);
		} catch (HibernateException he) {
			logger.warn("update(object = " + object + ") - failed", he);
			this.current().clear();
			throw new DAOException("errors.session.update", he);
		}
	}

	/**
	 * @see net.smartlab.web.DataAccessObject#list(net.smartlab.web.DataAccessObject.SearchInfo)
	 */
	public Collection list(SearchInfo info) throws DAOException {
		if (logger.isDebugEnabled()) {
			logger.debug("list(info = " + info + ") - start");
		}
		Criteria criteria = this.createCriteria(info);
		try {
			return criteria.list();
		} catch (HibernateException he) {
			logger.warn("list(info = " + info + ") - failed", he);
			throw new DAOException("errors.session.search", he);
		}
	}

	/**
	 * TODO documentation
	 * @param info
	 * @return
	 * @throws DAOException
	 */
	public Collection page(SearchInfo info) throws DAOException {
		if (logger.isDebugEnabled()) {
			logger.debug("page(info = " + info + ") - start");
		}
		Criteria criteria = this.createCriteria(info);
		try {
			return new Paginator(criteria);
		} catch (HibernateException he) {
			logger.warn("page(info = " + info + ") - error", he);
			throw new DAOException("errors.session.paging", he);
		}
	}

	/**
	 * TODO documentation
	 * @return
	 */
	public abstract Class getMappedClass();

	/**
	 * TODO documentation
	 * @param key
	 * @return
	 */
	public Serializable convertKey(Serializable key) {
		if (key != null) {
			if (key instanceof String) {
				if (key.equals("0")) {
					return null;
				} else {
					return new Long((String)key);
				}
			} else if (key instanceof Number) {
				if (((Number)key).longValue() == 0) {
					return null;
				} else {
					return new Long(((Number)key).toString());
				}
			}
		}
		return key;
	}

	/**
	 * TODO documentation
	 * @param info
	 * @return
	 * @throws DAOException
	 */
	public Criteria createCriteria(SearchInfo info) throws DAOException {
		if (logger.isDebugEnabled()) {
			logger.debug("createCriteria(info = " + info + ") - start");
		}
		Session session = this.current();
		Criteria criteria = session.createCriteria(this.getMappedClass());
		if (info.isUnion()) {
			// TODO the filters are in OR
		}
		Iterator filters = info.getFilters().iterator();
		while (filters.hasNext()) {
			if (logger.isTraceEnabled()) {
				logger.trace("createCriteria(info = " + info + ") - parsing filter `" + filters + "`");
			}
			Filter filter = (Filter)filters.next();
			switch (filter.getCondition()) {
				case SearchInfo.EQUALS:
					for (int i = 0; i < filter.getValues().length; i++) {
						criteria.add(Expression.eq(filter.getColumn(), filter.getValue(i)));
					}
					break;
				case SearchInfo.GREATER:
					for (int i = 0; i < filter.getValues().length; i++) {
						criteria.add(Expression.gt(filter.getColumn(), filter.getValue(i)));
					}
					break;
				case SearchInfo.GREATER_EQUALS:
					for (int i = 0; i < filter.getValues().length; i++) {
						criteria.add(Expression.ge(filter.getColumn(), filter.getValue(i)));
					}
					break;
				case SearchInfo.LESSER:
					for (int i = 0; i < filter.getValues().length; i++) {
						criteria.add(Expression.le(filter.getColumn(), filter.getValue(i)));
					}
					break;
				case SearchInfo.LESSER_EQUALS:
					for (int i = 0; i < filter.getValues().length; i++) {
						criteria.add(Expression.ge(filter.getColumn(), filter.getValue(i)));
					}
					break;
				case SearchInfo.NOT_EQUALS:
					for (int i = 0; i < filter.getValues().length; i++) {
						criteria.add(Expression.not(Expression.eq(filter.getColumn(), filter.getValue(i))));
					}
					break;
				case SearchInfo.BETWEEN:
					criteria.add(Expression.between(filter.getColumn(), filter.getValue(0), filter.getValue(1)));
					break;
				case SearchInfo.LIKE:
					for (int i = 0; i < filter.getValues().length; i++) {
						criteria.add(Expression.like(filter.getColumn(), filter));
					}
					break;
			}
		}
		if (info.getOrder() != null && info.getOrder().length() > 0) {
			if (info.isDescendant()) {
				criteria.addOrder(Order.desc(info.getOrder()));
			} else {
				criteria.addOrder(Order.asc(info.getOrder()));
			}
		}
		return criteria;
	}

	/**
	 * TODO documentation
	 * @throws DAOException
	 */
	public void begin() throws DAOException {
		if (logger.isDebugEnabled()) {
			logger.debug("begin() - start");
		}
		if (BusinessObjectFactory.transaction.get() == null) {
			try {
				BusinessObjectFactory.transaction.set(this.current().beginTransaction());
			} catch (HibernateException he) {
				logger.error("begin() - error", he);
				throw new DAOException("errors.session.begin", he);
			}
		} else {
			logger.warn("begin() - transaction already started");
		}
	}

	/**
	 * TODO documentation
	 * @throws DAOException
	 */
	public void commit() throws DAOException {
		if (logger.isDebugEnabled()) {
			logger.debug("commit() - start");
		}
		Transaction transaction = (Transaction)BusinessObjectFactory.transaction.get();
		try {
			transaction.commit();
		} catch (HibernateException he) {
			logger.error("commit() - error", he);
			throw new DAOException("errors.session.commit", he);
		} catch (NullPointerException npe) {
			logger.warn("commit() - transaction never started");
			throw new DAOException("errors.session.commit");
		} finally {
			BusinessObjectFactory.transaction.set(null);
		}
	}

	/**
	 * TODO documentation
	 * @throws DAOException
	 */
	public void rollback() throws DAOException {
		if (logger.isDebugEnabled()) {
			logger.debug("rollback() - start");
		}
		Transaction transaction = (Transaction)BusinessObjectFactory.transaction.get();
		try {
			transaction.rollback();
		} catch (HibernateException he) {
			logger.error("rollback() - error", he);
			throw new DAOException("errors.session.rollback", he);
		} catch (NullPointerException npe) {
			logger.warn("rollback() - transaction never started", npe);
		} finally {
			BusinessObjectFactory.transaction.set(null);
		}
	}


	/**
	 * TODO documentation
	 */
	public class Paginator extends net.smartlab.web.page.Paginator implements Collection {

		/**
		 * Logger for this class
		 */
		private final Log logger = LogFactory.getLog(Paginator.class);

		/**
		 * References the Hibernate query or criteria passed to the constructor.
		 * It cannot be null.
		 */
		private Object statement;


		/**
		 * Constructs an instance on an Hibernate criteria with an unlimited
		 * number of items per page and an unlimited number of pages displayed.
		 * 
		 * @param criteria the Hibernate criteria whose results must be
		 *        paginated.
		 * @throws HibernateException if something wrong occurs accessing the
		 *         database.
		 */
		public Paginator(Criteria criteria) throws HibernateException {
			this(criteria, Paginator.UNLIMITED_ITEMS, Paginator.UNLIMITED_PAGES);
		}

		/**
		 * Constructs an instance on an Hibernate criteria with the specified
		 * number of elements per page and the specified number of pages
		 * displayed.
		 * 
		 * @param criteria the Hibernate criteria whose results must be
		 *        paginated.
		 * @param size the number of elements per page.
		 * @param pages the number of pages to be displayed for quick
		 *        navigation.
		 * @throws HibernateException if something wrong occurs accessing the
		 *         database.
		 */
		public Paginator(Criteria criteria, int size, int pages) throws HibernateException {
			super(size, pages);
			this.statement = criteria;
			List temp = new ArrayList();
			Iterator orderings = ((CriteriaImpl)criteria).iterateOrderings();
			while (orderings.hasNext()) {
				Order ordering = ((CriteriaImpl.OrderEntry)orderings.next()).getOrder();
				orderings.remove();
				temp.add(ordering);
			}
			criteria.setProjection(Projections.rowCount());
			this.setCount(((Integer)criteria.uniqueResult()).intValue());
			criteria.setProjection(null);
			criteria.setResultTransformer(Criteria.ROOT_ENTITY);
			orderings = temp.iterator();
			while (orderings.hasNext()) {
				criteria.addOrder((Order)orderings.next());
			}
			this.setSize(size);
		}

		/**
		 * Constructs an instance on an Hibernate query with an unlimited number
		 * of items per page and an unlimited number of pages displayed.
		 * 
		 * @param query the Hibernate query whose results must be paginated.
		 * @throws DAOException if something wrong occurs accessing the
		 *         underling data layer.
		 * @throws HibernateException if something wrong occurs accessing the
		 *         database.
		 */
		public Paginator(Query query) throws DAOException, HibernateException {
			this(query, Paginator.UNLIMITED_ITEMS, Paginator.UNLIMITED_PAGES);
		}

		/**
		 * Constructs an instance on an Hibernate query with the specified
		 * number of elements per page and the specified number of pages
		 * displayed. FIXME the computation of rowCount is still slow on long
		 * lists.
		 * 
		 * @param query the Hibernate query whose results must be paginated.
		 * @param size the number of elements per page.
		 * @param pages the number of pages to be displayed for quick
		 *        navigation.
		 * @throws DAOException if something wrong occurs accessing the
		 *         underling data layer.
		 * @throws HibernateException if something wrong occurs accessing the
		 *         database.
		 */
		public Paginator(Query query, int size, int pages) throws DAOException, HibernateException {
			super(size, pages);
			this.statement = query;
			this.setCount(query.list().size());
			this.setSize(size);
		}

		/**
		 * @see net.smartlab.web.page.Paginator#setPage(int)
		 */
		public void setPage(int page) {
			if (logger.isDebugEnabled()) {
				logger.debug("setPage(page = " + page + ") - start");
			}
			if (statement instanceof Criteria) {
				((Criteria)this.statement).setFirstResult((page - 1) * super.getPageSize());
			} else {
				((Query)this.statement).setFirstResult((page - 1) * super.getPageSize());
			}
			super.setPage(page);
		}

		/**
		 * TODO documentation
		 * @param size
		 */
		public void setSize(int size) {
			if (logger.isDebugEnabled()) {
				logger.debug("setSize(size = " + size + ") - start");
			}
			if (size != UNLIMITED_ITEMS) {
				super.setPageSize(size);
				if (statement instanceof Criteria) {
					((Criteria)this.statement).setMaxResults(size);
				} else {
					((Query)this.statement).setMaxResults(size);
				}
			} else {
				super.setPageSize(this.getCount());
			}
		}

		/**
		 * @see net.smartlab.web.page.Paginator#setArray()
		 */
		protected void setArray() {
			if (logger.isDebugEnabled()) {
				logger.debug("setArray() - start");
			}
			try {
				Iterator elements = null;
				if (statement instanceof Criteria) {
					elements = ((Criteria)this.statement).list().iterator();
				} else {
					elements = ((Query)this.statement).list().iterator();
				}
				for (int i = 0; i < array.length && elements.hasNext(); i++) {
					array[i] = elements.next();
				}
			} catch (HibernateException he) {
				logger.error("setArray() - error", he);
			}
		}

		/**
		 * @see java.util.Collection#add(java.lang.Object)
		 */
		public boolean add(Object obj) {
			throw new UnsupportedOperationException();
		}

		/**
		 * @see java.util.Collection#addAll(java.util.Collection)
		 */
		public boolean addAll(Collection collection) {
			throw new UnsupportedOperationException();
		}

		/**
		 * @see java.util.Collection#clear()
		 */
		public void clear() {
			throw new UnsupportedOperationException();
		}

		/**
		 * @see java.util.Collection#contains(java.lang.Object)
		 */
		public boolean contains(Object item) {
			if (logger.isDebugEnabled()) {
				logger.debug("contains(item = " + item + ") - start");
			}
			for (int i = 0; i < array.length; i++) {
				if (array[i].equals(item)) {
					return true;
				}
			}
			return false;
		}

		/**
		 * @see java.util.Collection#containsAll(java.util.Collection)
		 */
		public boolean containsAll(Collection item) {
			throw new UnsupportedOperationException();
		}

		/**
		 * @see java.util.Collection#isEmpty()
		 */
		public boolean isEmpty() {
			return super.getCount() == 0;
		}

		/**
		 * @see java.util.Collection#iterator()
		 */
		public Iterator iterator() {
			return this;
		}

		/**
		 * @see java.util.Collection#remove(java.lang.Object)
		 */
		public boolean remove(Object obj) {
			throw new UnsupportedOperationException();
		}

		/**
		 * @see java.util.Collection#removeAll(java.util.Collection)
		 */
		public boolean removeAll(Collection collection) {
			throw new UnsupportedOperationException();
		}

		/**
		 * @see java.util.Collection#retainAll(java.util.Collection)
		 */
		public boolean retainAll(Collection collection) {
			throw new UnsupportedOperationException();
		}

		/**
		 * @see java.util.Collection#size()
		 */
		public int size() {
			return super.getCount();
		}

		/**
		 * @see java.util.Collection#toArray()
		 */
		public Object[] toArray() {
			return array;
		}

		/**
		 * @see java.util.Collection#toArray(java.lang.Object[])
		 */
		public Object[] toArray(Object[] array) {
			return this.array;
		}
	}
}
